/*
HASH.C

Routines that manage hash lists; each named hash list is associated with
a particular struct (ie address) in memory.
*/

/* Each item within a hash is a simple key/value pair */
typedef struct {
	int keylen, valuelen;
	char *key;
	char *value;
} item_t;

/* Each allocated hash array contains the key/value pairs themselves,
   along with the address of the owner, the name of the hash, and
   statistics relating the allocated key/value pairs. */
typedef struct {
	item_t** entries;
	void* owner;
	char name[128];
	int num;
	int max;
	int array;
	int inuse;
} HASH;

/* Maximum allowed hashes */
#define MAX_HASHES 65536


/* HASH_GLOBAL can be used as the owner of lists which apply throughout
   the program's execution. */
#define HASH_GLOBAL		(NULL)

/* Used as a key value, means 'any' for example, doing an insert without
specifying a key */
#define HASH_PARAM		void*
#define HASH_ANY		((HASH_PARAM) -1)


/* These three definitions should be passed as the 'type' parameter
   to functions that require it */
#define HASH_STRING		("%s")
#define HASH_INT		("%d")
#define HASH_PTR		("%p")


/* Routines to create and destroy hashes */
void	HashInit(HASH* h[], int* nh);
HASH*	HashAlloc(void* owner, char* name, int size);
int		HashFree(void* owner, char* name);
HASH*	HashCopy(void* owner, char* name, void* orig_owner, void* orig_name);
void	HashFreeAll(void);
void	HashFreeAssoc(void* owner);

/* Routines to insert, select, or delete key/value pairs */
void*	HashInsert(void* owner, char* name, void* key, void* value, char* type);
void*	HashSelect(void* owner, char* name, void* key, char* type);
int		HashDelete(void* owner, char* name, void* key);
int		HashDeleteValue(void* owner, char* name, void* value, char* type);
int		HashDeleteAll(void* owner, char* name);

/* Returns the HASH* to the specified hash */
HASH*	HashGet(void* owner, char* name);
/* Counts the number of key/value pairs in the specified hash */
int		HashCount(void* owner, char* name);
/* Moves hash elements around */
int		HashMove(void* owner, char* name, int from, int to);

/* Routines to traverse an array or find a particular element */
void*   HashReset(void* owner, char* name, char* type);
void*   HashNext(void* owner, char* name, char* type);
void*   HashFind(void* owner, char* name, char* str, int offset);
HASH*	HashExists(void* owner, char* name);
/* These two return -1 on error */
void*   HashExistsKey(void* owner, char* name, void* key);
void*   HashExistsValue(void* owner, char* name, void* value, char* type);

/* Routines that deal with other sets */
int		HashExistsSubhash(void* owner, char*name, void* subhash_owner, char* subhash_name);
int		HashInsertHash(void* owner, char* name, void* other, char* other_name);


/* Aliased functions to make life easier */
#define HashSelectText(x,y,k) ((char*)HashSelect(x,y,k,HASH_STRING))
#define HashSelectInt(x,y,k) ((int)HashSelect(x,y,k,HASH_INT))
#define HashSelectPtr(x,y,k) HashSelect(x,y,k,HASH_PTR)

#define HashInsertText(x,y,k,v) ((char*)HashInsert(x,y,k,v,HASH_STRING))
#define HashInsertInt(x,y,k,v) ((int)HashInsert(x,y,k,(HASH_PARAM)(v),HASH_INT))
#define HashInsertPtr(x,y,k,v) HashInsert(x,y,k,v,HASH_PTR)

#define HST HashSelectText
#define HSI HashSelectInt
#define HSP HashSelectPtr

#define HIT HashInsertText
#define HII HashInsertInt
#define HIP HashInsertPtr

#define HSTG(y,k) HST(HASH_GLOBAL,y,k)
#define HSIG(y,k) HSI(HASH_GLOBAL,y,k)
#define HSPG(y,k) HSP(HASH_GLOBAL,y,k)

#define HITG(y,k,v) HIT(HASH_GLOBAL,y,k,v)
#define HIIG(y,k,v) HII(HASH_GLOBAL,y,k,v)
#define HIPG(y,k,v) HIP(HASH_GLOBAL,y,k,v)


